<?php

namespace BS\ChatGPTQuickReplyBot\Service;

use BS\ChatGPTBots\Repository\Message;
use BS\ChatGPTBots\Response;
use BS\ChatGPTQuickReplyBot\Support\BotHelper;
use Orhanerday\OpenAi\OpenAi;
use XF\App;
use XF\Entity\Post;
use XF\Entity\Thread;
use XF\Entity\User;
use XF\Service\AbstractService;

class ReplyOnPost extends AbstractService
{
    /**
     * @var \XF\Entity\Thread
     */
    private Thread $thread;
    /**
     * @var \XF\Entity\User
     */
    private User $user;

    protected OpenAi $api;
    private Message $messageRepo;

    private array $replies = [];

    private array $messages = [];

    public function __construct(
        App $app,
        Thread $thread,
        User $user
    ) {
        parent::__construct($app);
        $this->setThread($thread);
        $this->setUser($user);
        $this->api = $app->container('chatGPT');
        $this->messageRepo = $this->repository('BS\ChatGPTBots:Message');
    }

    public function postReplies()
    {
        if (empty($this->replies)) {
            return null;
        }

        /** @var \XF\Service\Thread\Replier $replier */
        $replier = $this->service('XF:Thread\Replier', $this->thread);
        $replier->setMessage($this->getFinalMessage());

        if ($this->app->options()->bsChatGPTAutoresponderSpamCheck) {
            $replier->logIp(false);
        } else {
            $replier->setIsAutomated();
        }

        $post = $replier->save();
        $replier->sendNotifications();

        $this->clearMessagesMemory();

        return $post;
    }

    protected function getFinalMessage(): string
    {
        return implode(str_repeat(PHP_EOL, 2), $this->replies);
    }

    public function getRepliesOnQuotes(
        Post $post,
        array $quotes
    ): array {
        $em = $this->em();

        if (empty($this->messages)) {
            // give bot context of the post it's replying to
            $this->messages = $this->buildMessagesForPostReply($post);
        }

        foreach ($quotes as $quote) {
            /** @var Post $quotedPost */
            $quotedPost = $em->find('XF:Post', $quote['post_id']);
            if (! $quotedPost) {
                continue;
            }

            $this->addQuoteToMessages($quote);
            $reply = $this->getReplyForMessages();
            // Add bot replies to memory
            $this->messages[] = $this->messageRepo->wrapMessage($reply, 'assistant');
            $this->pushReplyWithQuote($post, $quote, $reply);
        }

        return $this->replies;
    }

    protected function pushReplyWithQuote(
        Post $post,
        array $quote,
        string $reply
    ) {
        $replyQuote = $this->makePartialQuote($post, $quote['message']);
        $reply ??= $this->getReplyForMessages();

        $this->replies[] = $replyQuote . PHP_EOL . $reply;
    }

    public function getReplyForMessages(): string
    {
        if (empty($this->messages)) {
            return '';
        }

        $options = \XF::options();
        return Response::getReplyWithLogErrors($this->api, [
            'model'             => $options->bsChatGPTAutoresponderModel,
            'messages'          => $this->messages,
            'temperature'       => (float)$options->bsChatGPTQuickReplyTemp,
            'frequency_penalty' => 0,
            'presence_penalty'  => 0,
        ]);
    }

    protected function buildMessagesForPostReply(Post $post): array {
        return [
            BotHelper::getSystemForThreadReply($this->thread),
            ...$this->messageRepo->fetchMessagesFromThread(
                $this->thread,
                $post->position,
                $this->user,
                true,
                max(0, $post->position - 10)
            )
        ];
    }

    protected function addQuoteToMessages(array $quote): void
    {
        $this->messages[] = $this->messageRepo->wrapMessage($quote['content'], 'assistant');
        $this->messages[] = $this->messageRepo->wrapMessage($quote['message']);
    }

    public function clearMessagesMemory(): void
    {
        $this->messages = [];
    }

    protected function makePartialQuote(
        Post $post,
        string $content
    ): string {
        return "[QUOTE=\"{$post->username}, post: {$post->post_id}, member: {$post->user_id}\"]{$content}[/QUOTE]";
    }

    /**
     * @param  \XF\Entity\Thread  $thread
     */
    public function setThread(Thread $thread): void
    {
        $this->thread = $thread;
    }

    /**
     * @param  \XF\Entity\User  $user
     */
    public function setUser(User $user): void
    {
        $this->user = $user;
    }
}